/* diversity.c 

  (C) Norbert H. Doerry 2014
  November 2014

  diversity.exe filename.xml -d


This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.



*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "diversity.h"
#include "xml.h"

#define VERSION "1.0"


int calc_stats(STATISTICS *s);

int main(int argc, char **argv)
{
 DIVERSITY d;
 FILE *xin,*cin,*sout,*dout;
 int exit_flag;

 exit_flag = EXIT_SUCCESS;

if (Diversity_initialize(&d) != EXIT_SUCCESS)
 {
  fprintf(stderr," *** ERROR in Diversity_initialize()\n");
  exit(EXIT_FAILURE);
 }

if (Diversity_read_command_line(&d,argc, argv) != EXIT_SUCCESS)
 {
  fprintf(stderr," *** ERROR in Diversity_read_command_line()\n");
  exit(EXIT_FAILURE);
 } 

if (d.debug != 0) {printf("."); fflush(stdout);}   /* first period */

/* read xml file */

if (d.input_xml_filename == (char *) NULL)
 {
  fprintf(stderr," *** ERROR .xml input file not specified on command line\n");
  exit(EXIT_FAILURE);
 } 

 xin = fopen(d.input_xml_filename,"r");

 if (xin == (FILE *) NULL)
 {
  fprintf(stderr," *** ERROR unable to open file %s for reading\n", d.input_xml_filename);
  exit(EXIT_FAILURE);
 } 

 
if (Diversity_read_input_xml_file(&d, xin) != EXIT_SUCCESS)
 {
  fprintf(stderr," *** ERROR in Diversity_read_input_xml_file()\n");
  exit(EXIT_FAILURE);
 } 

fclose(xin);

if (d.debug != 0) {printf("."); fflush(stdout);}   /* second period */

/* read data file */

if (d.input_data_filename == (char *) NULL)
 {
  fprintf(stderr," *** ERROR input data file not specified in %s\n",d.input_xml_filename);
  exit(EXIT_FAILURE);
 } 

 cin = fopen(d.input_data_filename,"r");

 if (cin == (FILE *) NULL)
 {
  fprintf(stderr," *** ERROR unable to open file %s for reading\n", d.input_data_filename);
  exit(EXIT_FAILURE);
 } 


if (Diversity_read_input_csv_file(&d, cin) != EXIT_SUCCESS)
 {
  fprintf(stderr," *** ERROR in Diversity_read_input_csv_file()\n");
  exit(EXIT_FAILURE);
 } 

fclose(cin);

if (d.debug != 0) {printf("."); fflush(stdout);}   /* third period */

if (Diversity_sort_data(&d) != EXIT_SUCCESS)
 {
  fprintf(stderr," *** ERROR in Diversity_sort_data()\n");
  exit(EXIT_FAILURE);
 } 

if (d.debug != 0) {printf("."); fflush(stdout);}  /* fourth Period */

if (Diversity_count_diversity_fields(&d) != EXIT_SUCCESS)
 {
  fprintf(stderr," *** ERROR in Diversity_count_diversity_fields()\n");
  exit(EXIT_FAILURE);
 } 

if (d.debug != 0) {printf("."); fflush(stdout);}   /* fifth period */

if (Diversity_id_points(&d) != EXIT_SUCCESS)
 {
  fprintf(stderr," *** ERROR in Diversity_id_points()\n");
 } 

if (d.debug != 0) {printf("."); fflush(stdout);}    /* sixth period */
  
if (Diversity_calculate_statistics(&d) != EXIT_SUCCESS)
 {
  fprintf(stderr," *** ERROR in Diversity_calculate_statistics()\n");
  exit(EXIT_FAILURE);
 } 

if (d.debug != 0) {printf("."); fflush(stdout);}   /* seventh period */

/* write statistics file if specified */

if (d.output_statistics_filename != (char *) NULL)
 {

  sout = fopen(d.output_statistics_filename,"w");

  if (sout == (FILE *) NULL)
   {
    fprintf(stderr," *** ERROR unable to open output statistics file %s for writing\n", d.output_statistics_filename);
    exit_flag = EXIT_FAILURE;
   } 
  else
   {
    if (Diversity_write_statistics_file(&d,sout) != EXIT_SUCCESS)
     {
      fprintf(stderr," *** ERROR in Diversity_write_statics_file()\n");
      exit_flag = EXIT_FAILURE;
     }
    fclose(sout);
   } 
  if (d.debug != 0) {printf("+"); fflush(stdout);}
 }

if (d.debug != 0) {printf("."); fflush(stdout);} /* eighth period */

/* write data file if specified */

if (d.output_data_filename != (char *) NULL)
 {
  dout = fopen(d.output_data_filename,"w");

  if (dout == (FILE *) NULL)
   {
    fprintf(stderr," *** ERROR unable to open output data file %s for writing\n", d.output_data_filename);
    exit_flag = EXIT_FAILURE;
   } 
  else
   {
    if (Diverstiy_write_output_data_file(&d,dout) != EXIT_SUCCESS)
     {
      fprintf(stderr," *** ERROR in Diversity_write_output_data_file()\n");
      exit_flag = EXIT_FAILURE;
     } 
    fclose(dout);
   }
  if (d.debug != 0) {printf("+"); fflush(stdout);}
 }

 if (d.debug != 0) {printf(".\n"); fflush(stdout);}  /* ninth period */
 return exit_flag;

}






int Diversity_initialize(DIVERSITY *d)
{
 
 d->input_xml_filename = (char *) NULL;
 d->input_data_filename = (char *) NULL;

 d->output_statistics_filename = (char *) NULL;
 d->output_data_filename = (char *) NULL;

 d->name = (char *) NULL;
 d->preference = PREFERENCE_LOW;
 d->min_nbr_fields = 3;
 
 d->data = (DATA_ROW *) NULL;
 d->pdata = (DATA_ROW **) NULL;
 d->nbr_data_row = 0;
 d->nbr_pdata_row = 0;

 d->row_dm = 0;

 d->metric_base_sum = 0;

 d->sort_field = (FIELD *) NULL;
 d->statistics_field = (FIELD *) NULL;
 d->output_field = (FIELD *) NULL;
 d->diversity_field = (FIELD *) NULL;
 
 d->output_diversity_metric_flag = 0;

 d->debug = 0;
 
 return EXIT_SUCCESS;
}

int Diversity_read_command_line(DIVERSITY *d,int argc, char **argv)
{
 int i;

 for (i = 1 ; i < argc ; i++)
  {
   if (argv[i][0] == '-')
    {
     if (argv[i][1] == 'd' || argv[i][1] == 'D')
      d->debug = 1;

     else
      {
       fprintf(stderr," *** WARNING: unknown command line switch (%s)\n", argv[i]);
      }

    }
   else if (d->input_xml_filename == (char *) NULL)
    {
     d->input_xml_filename = argv[i];
  
    }
   else
    {
     fprintf(stderr," *** ERROR in Diverstiy_read_command_line(): multiple xml files on command line (%s, %s)\n",
		d->input_xml_filename,argv[i]);
     return EXIT_FAILURE;
    }
  }

 return EXIT_SUCCESS;
}

int Diversity_read_input_xml_file(DIVERSITY *d, FILE *xin)
{
 XMLENTITY *xml, *x1, *x2;
 int errflag;
 double val;
 int i;
 FIELD *fd;

 xml = readfile_XMLENTITY(xin,&errflag);

 if (xml == (XMLENTITY *) NULL) return EXIT_FAILURE;
 if (errflag != ERROR_NONE) return EXIT_FAILURE;

 if (strcmp(xml->name,"DIVERSITY") != 0 && strcmp(xml->name,"diversity") != 0)
  {
   fprintf(stderr," *** ERROR: Root xml entity not <DIVERSITY> (%s)\n", xml->name);
   return EXIT_FAILURE; 
  }
 
 for (x1 = xml->child ; x1 != (XMLENTITY *) NULL ; x1 = x1->next)
  {
   if (strcmp(x1->name,"NAME") == 0 || strcmp(x1->name,"name") == 0)
    {
     if (d->name != (char *) NULL)
      {
       fprintf(stderr," *** ERROR: NAME defined multiple times (%s)(%s)\n",d->name,x1->s);
       return EXIT_FAILURE;
      }
 
     d->name = copystring(x1->s);
    }
   else if (strcmp(x1->name,"COMMENT") == 0 || strcmp(x1->name,"comment") == 0)
    {

    }
   else if (strcmp(x1->name,"MIN_NBR_FIELDS") == 0 || strcmp(x1->name,"min_nbr_fields") == 0)
    {
     /* this value only applies if it is greater than the number of the lowest field number
         specified in one of the FIELD commands, or less than the number of fields 
         in the first line of the .csv file */

      val = strtod(x1->s,(char **) NULL);

      if (val > 1)
	d->min_nbr_fields = val;
       
    }
   else if (strcmp(x1->name,"SOURCE_FILE") == 0 || strcmp(x1->name,"source_file") == 0)
    {   

     if (d->input_data_filename != (char *) NULL)
      {
       fprintf(stderr," *** ERROR: SOURCE_FILE defined multiple times (%s)(%s)\n",
		d->input_data_filename,x1->s);
       return EXIT_FAILURE;
      }


      d->input_data_filename = copystring(x1->s);
    }
   else if (strcmp(x1->name,"SORT_FIELD") == 0 || strcmp(x1->name,"sort_field") == 0)
    {   
      for (x2 = x1->child ; x2 != (XMLENTITY *) NULL ; x2 = x2->next)
       {
         if (strcmp(x2->name,"COMMENT") == 0 || strcmp(x2->name,"comment") == 0)
          {
		/* ignore Comments */
          } 
         else if (strcmp(x2->name,"NAME") == 0 || strcmp(x2->name,"name") == 0)
          {
	   if (d->sort_field != (FIELD *) NULL)
            {
              fprintf(stderr," *** ERROR: SORT_FIELD defined multiple times (%s)(%s)\n",
		d->sort_field->name,x2->s);
              return EXIT_FAILURE;
            }

           d->sort_field = (FIELD *) calloc((size_t) 1, sizeof(FIELD));
           if (d->sort_field == (FIELD *) NULL)
            {
             fprintf(stderr,
              " *** ERROR:  Out of memory in Diversity_read_input_xml_file_1()\n");
             exit(EXIT_FAILURE);
            }
           d->sort_field->name = x2->s;
           
	   /* set the column to -1 for now */
 
           d->sort_field->col = -1;

	   
           /* set the rest of the data elements */

           d->sort_field->next = (FIELD *) NULL;
           d->sort_field->min_nbr_options = 0;
           d->sort_field->min_nbr_configs_per_option = 0;
	   d->sort_field->st = (STATISTICS *) NULL;
 
          } 

         else if (strcmp(x2->name,"PREFERENCE") == 0 || strcmp(x2->name,"preference") == 0)
          {
	   if (strcmp(x2->s,"LOW") == 0 || strcmp(x2->s,"low") == 0)
            d->preference = PREFERENCE_LOW;
           else if (strcmp(x2->s,"HIGH") == 0 || strcmp(x2->s,"high") == 0)
            d->preference = PREFERENCE_HIGH;
           else
            {
             fprintf(stderr," *** WARNING: Unknown PREFERENCE (%s) using default LOW\n",x2->s);
  	     d->preference = PREFERENCE_LOW;
            } 	
          } 
         else
          {
            fprintf(stderr," *** WARNING: Unknown Entity <%s>\n", x2->name);
          }
     
       }
    }
   else if (strcmp(x1->name,"STATISTICS_FIELDS") == 0 || 
	    strcmp(x1->name,"statistics_fields") == 0)
    {   
      for (x2 = x1->child ; x2 != (XMLENTITY *) NULL ; x2 = x2->next)
       {
         if (strcmp(x2->name,"COMMENT") == 0 || strcmp(x2->name,"comment") == 0)
          {
		/* ignore Comments */
          } 

         else if (strcmp(x2->name,"NAME") == 0 || strcmp(x2->name,"name") == 0)
          {
           fd = (FIELD *) calloc((size_t) 1, sizeof(FIELD));
           if (fd == (FIELD *) NULL)
            {
             fprintf(stderr,
              " *** ERROR:  Out of memory in Diversity_read_input_xml_file_2()\n");
             exit(EXIT_FAILURE);
            }
           fd->name = x2->s;
           
	   /* set the column to -1 for now */
 
           fd->col = -1;

           /* allocate the statistics structure */

           fd->st = (STATISTICS *) calloc((size_t) 1, sizeof(STATISTICS));
           if (fd->st == (STATISTICS *) NULL)
            {
             fprintf(stderr,
              " *** ERROR:  Out of memory in Diversity_read_input_xml_file_2a()\n");
             exit(EXIT_FAILURE);
            }

	   fd->st->mean = 0.0;
           fd->st->median = 0.0;
           fd->st->standard_deviation = 0.0;
           fd->st->percentile_5 = 0.0;
           fd->st->percentile_95 = 0.0;
    	   fd->st->weighted_mean = 0.0;
           fd->st->min = 0.0;
	   fd->st->max = 0.0;


	   fd->st->n = 0;
	   fd->st->val = (double *) NULL;
           fd->st->weight = (double *) NULL;
	   
           /* set the rest of the data elements */

           fd->next = d->statistics_field;
           fd->min_nbr_options = 0;
           fd->min_nbr_configs_per_option = 0;
           d->statistics_field = fd;
		
          } 
         else if (strcmp(x2->name,"FILENAME") == 0 || strcmp(x2->name,"filename") == 0)
          {
           if (d->output_statistics_filename != (char *) NULL)
            {
             fprintf(stderr,
                " *** ERROR: STATISTICS_FIELDS FILENAME defined multiple times (%s)(%s)\n",
		d->output_statistics_filename,x2->s);
             return EXIT_FAILURE;
            }
 
           d->output_statistics_filename = copystring(x2->s);
		
          } 
         else
          {
            fprintf(stderr," *** WARNING: Unknown Entity <%s>\n", x2->name);
          }
     
       }
    }

   else if (strcmp(x1->name,"OUTPUT_FIELDS") == 0 || strcmp(x1->name,"output_fields") == 0)
    {
       for (x2 = x1->child ; x2 != (XMLENTITY *) NULL ; x2 = x2->next)
        {
         if (strcmp(x2->name,"COMMENT") == 0 || strcmp(x2->name,"comment") == 0)
          {
		/* ignore Comments */
          }      
         else if (strcmp(x2->name,"NAME") == 0 || strcmp(x2->name,"name") == 0)
          {
           fd = (FIELD *) calloc((size_t) 1, sizeof(FIELD));
           if (fd == (FIELD *) NULL)
            {
             fprintf(stderr,
              " *** ERROR:  Out of memory in Diversity_read_input_xml_file_3()\n");
             exit(EXIT_FAILURE);
            }
           fd->name = x2->s;
           
	   /* set the column to -1 for now */
 
           fd->col = -1;

	   
           /* set the rest of the data elements */
	   
	   fd->st = (STATISTICS *) NULL;	
           fd->next = d->output_field;
           fd->min_nbr_options = 0;
           fd->min_nbr_configs_per_option = 0;
           d->output_field = fd;
		
          } 
         else if (strcmp(x2->name,"FILENAME") == 0 || strcmp(x2->name,"filename") == 0)
          {
          if (d->output_data_filename != (char *) NULL)
            {
             fprintf(stderr,
                " *** ERROR: OUTPUT_FIELDS FILENAME defined multiple times (%s)(%s)\n",
		d->output_data_filename,x2->s);
             return EXIT_FAILURE;
            }
 
           d->output_data_filename = copystring(x2->s);
		
          }
         else if (strcmp(x2->name,"DIVERSITY_METRIC") == 0 || 
		  strcmp(x2->name,"diversity_metric") == 0) 
          {
           if (strcmp(x2->s,"ON") == 0 || strcmp(x2->s,"on") == 0)
            d->output_diversity_metric_flag = 1;
           else if (strcmp(x2->s,"OFF") == 0 || strcmp(x2->s,"off") == 0)
            d->output_diversity_metric_flag = 0;
           else
	    {
             fprintf(stderr," *** ERROR: unknown <DIVERSITY_METRIC> value = (%s)\n",x2->s);
             return EXIT_FAILURE;
            }
	  }
         else
          {
            fprintf(stderr," *** WARNING: Unknown Entity <%s>\n", x2->name);
          }

        }  
    }

   else if (strcmp(x1->name,"DIVERSITY_FIELD") == 0 || strcmp(x1->name,"diversity_field") == 0)
    { 
      fd = (FIELD *) calloc((size_t) 1, sizeof(FIELD));
      if (fd == (FIELD *) NULL)
       {
        fprintf(stderr,
         " *** ERROR:  Out of memory in Diversity_read_input_xml_file_4()\n");
        exit(EXIT_FAILURE);
       }
      fd->name = (char *) NULL;
      fd->col = -1;
      fd->next = d->diversity_field;
      d->diversity_field = fd;
      fd->min_nbr_options = 1;
      fd->min_nbr_configs_per_option = 1;
      fd->st = (STATISTICS *) NULL;
  
      for (x2 = x1->child ; x2 != (XMLENTITY *) NULL ; x2 = x2->next)
       {
         if (strcmp(x2->name,"COMMENT") == 0 || strcmp(x2->name,"comment") == 0)
          {
		/* ignore Comments */
          }      
         else if (strcmp(x2->name,"NAME") == 0 || strcmp(x2->name,"name") == 0)
          {
           fd->name = copystring(x2->s);
          } 
         else if (strcmp(x2->name,"MIN_NBR_OPTIONS") == 0 || 
			strcmp(x2->name,"min_nbr_options") == 0)
          {
	    fd->min_nbr_options = (int) strtod(x2->s,(char **) NULL);
            if (fd->min_nbr_options < 1) fd->min_nbr_options = 1;	
          } 
         else if (strcmp(x2->name,"MIN_NBR_CONFIGS_PER_OPTION") == 0 || 
			strcmp(x2->name,"min_nbr_configs_per_option") == 0)
          {
	    fd->min_nbr_configs_per_option = (int) strtod(x2->s,(char **) NULL);
            if (fd->min_nbr_configs_per_option < 1) fd->min_nbr_configs_per_option = 1;	
 	  } 
         else
          {
            fprintf(stderr," *** WARNING: Unknown Entity <%s>\n", x2->name);
          }


       }

     fd->metric_base = fd->min_nbr_options * fd->min_nbr_configs_per_option;
     d->metric_base_sum += fd->metric_base;
     
    }
   else
    {
      fprintf(stderr," *** WARNING: Unknown Entity <%s>\n", x1->name);

    }



  }

 return EXIT_SUCCESS;
}

int Diversity_read_input_csv_file(DIVERSITY *d, FILE *cin)
{

 DATA_ROW *dr,*pd;
 int ncnt,i; 

 FIELD *fd;
 
 dr = read_csv_file(cin,&ncnt);

 if (dr == (DATA_ROW *) NULL)
  return EXIT_FAILURE;
 
 d->data = dr;
 d->nbr_data_row = ncnt;


 if (d->debug != 0) {printf("+"); fflush(stdout);}  /* first + */

 /* set the column numbers for all the fields */

 /* row zero (first row) should have the field names */

 for (i = 0 ; i < d->data->n ; i++)
  {
    /*if (d->debug != 0) {printf("i = %d data->n = %d\n",i,d->data->n); 
    fflush(stdout);}*/


    if (strcmp(d->data->data[i],d->sort_field->name) == 0)
     {
      d->sort_field->col = i;
      /* if (d->debug != 0) {printf("SORT FIELD (%s) = %d\n",d->sort_field->name,i); 
       fflush(stdout);}*/
     }

    for (fd = d->statistics_field ; fd != (FIELD *) NULL ; fd = fd->next)
     {
       if (fd->col != -1) continue; 
       if (strcmp(d->data->data[i],fd->name) == 0)
        {
         fd->col = i;
         /*if (d->debug != 0) {printf("STATISTICS FIELD (%s) = %d\n",
		fd->name,i); fflush(stdout);}*/
        }
     }
    for (fd = d->output_field ; fd != (FIELD *) NULL ; fd = fd->next)
     { 
       if (fd->col != -1) continue; 
       if (strcmp(d->data->data[i],fd->name) == 0)
        {
         fd->col = i;
         /*if (d->debug != 0) {printf("OUTPUT FIELD (%s) = %d\n",
		fd->name,i); fflush(stdout);}*/
        }  
     }
    for (fd = d->diversity_field ; fd != (FIELD *) NULL ; fd = fd->next)
     {
       if (fd->col != -1) continue; 
       if (strcmp(d->data->data[i],fd->name) == 0)
        {
         fd->col = i;
         /*if (d->debug != 0) {printf("DIVERSITY FIELD (%s) = %d\n",
 		fd->name,i); fflush(stdout);}*/
  
        }
     }

  }

 if (d->debug != 0) {printf("+"); fflush(stdout);}  /* 2nd Plus */

 /* update the minimum number of columns and check for columns not found */

 if (d->min_nbr_fields > d->data->n)
  d->min_nbr_fields = d->data->n;  /* min number of fields can't be more than the number of fields in the first line */

 if (d->sort_field->col == -1)
  {
   fprintf(stderr," *** ERROR: SORT FIELD %s not found in .csv file\n",d->sort_field->name);
   return EXIT_FAILURE; 
  }
 else if (d->sort_field->col >= d->min_nbr_fields)
  d->min_nbr_fields = d->sort_field->col + 1;
  

 for (fd = d->statistics_field ; fd != (FIELD *) NULL ; fd = fd->next)
  {
   if (fd->col == -1)
    {
     fprintf(stderr," *** ERROR: STATISTICS FIELD %s not found in .csv file\n",fd->name);
     return EXIT_FAILURE; 
     }
    else if (fd->col >= d->min_nbr_fields)
     d->min_nbr_fields = fd->col + 1;
   }

 for (fd = d->output_field ; fd != (FIELD *) NULL ; fd = fd->next)
  {
   if (fd->col == -1)
    {
     fprintf(stderr," *** ERROR: OUTPUT FIELD %s not found in .csv file\n",fd->name);
     return EXIT_FAILURE; 
     }
    else if (fd->col >= d->min_nbr_fields)
     d->min_nbr_fields = fd->col + 1;
   }

  for (fd = d->diversity_field ; fd != (FIELD *) NULL ; fd = fd->next)
   {
    if (fd->col == -1)
     {
      fprintf(stderr," *** ERROR: DIVERSITY FIELD %s not found in .csv file\n",fd->name);
      return EXIT_FAILURE; 
      }
     else if (fd->col >= d->min_nbr_fields)
     d->min_nbr_fields = fd->col + 1;
   }

 if (d->debug != 0) {printf("+"); fflush(stdout);}  /* 3rd plus */


 
 /* create the table of pointers to each of the data rows */
  

 d->pdata = (DATA_ROW **) calloc((size_t) ncnt,sizeof(DATA_ROW *));

 if (d->pdata == (DATA_ROW **) NULL)
  {
   fprintf(stderr," *** ERROR:  Out of memory in Diverstiy_read_input_csv_file()\n");
   exit(EXIT_FAILURE);
  }

 for (i=0, pd = dr ; pd != (DATA_ROW *) NULL ; i++ , pd = pd->next);
  
 if (i != ncnt) 
  {
   fprintf(stderr," *** ERROR: Number of Data Row Elements calc error (%d != %d)\n",i,ncnt);
   return EXIT_FAILURE;
  }

 d->nbr_pdata_row = 0;

 /* loop through the data rows, ignoring the first row with headers */

 for (i=0, pd = dr->next ; pd != (DATA_ROW *) NULL ; pd = pd->next) 
  {
   /* ignore rows that do not have the minimum number of fields */
   
   if (pd->n < d->min_nbr_fields)
    continue;

   pd->fc = (FIELD_COUNT *) NULL;
   d->pdata[i] = pd;
   pd->select = 0;
   pd->diversity_score = 0.0;	
   d->nbr_pdata_row += 1;
   i++;

  }
 for ( ; i < ncnt ; i++)
  d->pdata[i] = (DATA_ROW *) NULL;

  if (d->debug != 0) {printf("+"); fflush(stdout);}  /* 4th plus */

 
 if (d->debug != 0)
  {
   if (d->sort_field == (FIELD *) NULL)
    printf("\nd->sort_field = NULL Pointer\n");
   else
    printf("\nSORT_FIELD %20s col = %d\n",d->sort_field->name,d->sort_field->col);

   for (fd = d->statistics_field; fd != (FIELD *) NULL ; fd = fd->next)
    printf("STATISTICS_FIELDS %20s col = %d\n",fd->name,fd->col);

   for (fd = d->output_field; fd != (FIELD *) NULL ; fd = fd->next)
    printf("OUTPUT_FIELDS %20s col = %d\n", fd->name,fd->col);

   for (fd = d->diversity_field; fd != (FIELD *) NULL ; fd = fd->next)
    printf("DIVERSITY_FIELD %20s col = %d\n", fd->name, fd->col);

  }
 return EXIT_SUCCESS;
}

int Diversity_sort_data(DIVERSITY *d)
{
  /* sort based on field number in d->sort_field->col  */

  int i,j,k;
  double x,y;
  DATA_ROW  *dr;

  if (d->sort_field == (FIELD *) NULL)
   {
    fprintf(stderr," *** ERROR:  d->sort_field is NULL pointer in Diversity_sort_data()\n");
    return EXIT_FAILURE;
   }

  if (d->sort_field->col < 0)
   {
    fprintf(stderr," *** ERROR:  d->sort_field->col is less than 0 (%d) in Diversity_sort_data()\n",
	d->sort_field->col);
    return EXIT_FAILURE;
   }

  if (d->preference != PREFERENCE_LOW && d->preference != PREFERENCE_HIGH)
   {
    fprintf(stderr," *** WARNING:  unknown d->preference flag (%d), assuming PREFERENCE_LOW\n", d->preference);

   }
   

  for (i=0 ; i < d->nbr_pdata_row ; i++)
   {
    if (d->pdata[i]->n <= d->sort_field->col)
     {
      fprintf(stderr," *** ERROR: line %d has too few columns (%d vice %d) in Diversity_sort_data()\n",
           i, d->pdata[i]->n, d->sort_field->col);
      return EXIT_FAILURE;
     }
   }

 if (d->debug != 0) {printf("+"); fflush(stdout);}  /* first + */

 if (d->debug != 0) {printf("\nd->preference = %d\n",d->preference); fflush(stdout);} 

  /* do the sort ... */  

  for (i=0; i < d->nbr_pdata_row ; i++)
   {

    x = strtod(d->pdata[i]->data[d->sort_field->col],(char **) NULL);

    /* if (d->debug != 0) {printf("i = %d d->nbr_pdata_row = %d x = %f\n",i,d->nbr_pdata_row,x); fflush(stdout);}*/ 

    for (j=i+1 ; j < d->nbr_pdata_row ; j++) 
     {

      y = strtod(d->pdata[j]->data[d->sort_field->col],(char **) NULL);
   
      /* if (d->debug != 0) {printf("j = %d  y = %f\n",j,y); fflush(stdout);} */

      if (d->preference == PREFERENCE_HIGH)
       {
        if (y > x)  /* swap the rows */
         {
          dr = d->pdata[i];
          d->pdata[i] = d->pdata[j];
          d->pdata[j] = dr;
	  x = y;
          /* if (d->debug != 0) {printf("swap x (%f) and y (%f)\n",x,y); fflush(stdout);}*/ 

         }
        
       }
      else
       {
        if (y < x)  /* swap the rows */
         {
          dr = d->pdata[i];
          d->pdata[i] = d->pdata[j];
          d->pdata[j] = dr;
	  x = y;
          /* if (d->debug != 0) {printf("swap x (%f) and y (%f)\n",x,y); fflush(stdout);} */ 

         }
        
       }
      


     }


   }


 if (d->debug != 0) 
  {

    for (i = 0 ; i < d->data->n - 1 ; i++)
     printf("%s,",d->data->data[i]);
    printf("%s\n",d->data->data[i]);

    for (j = 0 ; j < d->nbr_pdata_row ; j++)
     {
       for (i = 0 ; i < d->pdata[j]->n - 1 ; i++)
        printf("%s,",d->pdata[j]->data[i]);
       printf("%s\n",d->pdata[j]->data[i]);
     }
    fflush(stdout);
   } 



 return EXIT_SUCCESS;
}

int Diversity_count_diversity_fields(DIVERSITY *d)
{
 int i,j,k,tmp;

 int nbr_diversity_fields;

 DATA_ROW *prow;
 FIELD *df;
 FIELD_COUNT *fc,*fc1,*fc2;
 FIELD_COUNT_ELEMENT *fce,*fce1,*fce2;

 int *metric;


 /* allocte the field count linked lists based on d->diversity_field */
 

 nbr_diversity_fields = 0;
 for (df = d->diversity_field ; df != (FIELD *) NULL ; df = df->next)
  nbr_diversity_fields ++;
 
 /* set up the Field Count linked lists */

 for (i = 0 ; i < d->nbr_pdata_row ; i++)
  {
    d->pdata[i]->fc = (FIELD_COUNT *) NULL;
  
    for (df = d->diversity_field ; df != (FIELD *) NULL ; df = df->next)
     {
 
       fc = (FIELD_COUNT *) calloc((size_t) 1 , sizeof(FIELD_COUNT));
       if (fc == (FIELD_COUNT *) NULL)    
        {
         fprintf(stderr," *** ERROR:  Out of memory in Diversity_count_diversity_fields_1()\n");
         exit(EXIT_FAILURE);
        }
       fc->fld = df;
       fc->fce = (FIELD_COUNT_ELEMENT *) NULL;
       fc->Nbr_fce = 0;
       fc->Nbr_metric = 0;
       fc->next = d->pdata[i]->fc;
       d->pdata[i]->fc = fc;

      } 
  }
 
 /* start counting */


 if (d->debug)
  {
   printf("start counting\n");
   fflush(stdout);
  }

 /* set the count for each of the diversity fields for the first row to 1*/

 for (fc = d->pdata[0]->fc ; fc != (FIELD_COUNT *) NULL; fc = fc->next)
  {
   fce = (FIELD_COUNT_ELEMENT *) calloc((size_t) 1, sizeof(FIELD_COUNT_ELEMENT));
   if (fce == (FIELD_COUNT_ELEMENT *) NULL)    
    {
     fprintf(stderr," *** ERROR:  Out of memory in Diversity_count_diversity_fields_2()\n");
     exit(EXIT_FAILURE);
    }
   fce->next = (FIELD_COUNT_ELEMENT *) NULL;
   fce->nbr_configs = 1;
   fce->field_value = d->pdata[0]->data[fc->fld->col];
   fc->Nbr_fce = 1;
   fc->Nbr_metric = 1;  
   fc->fce = fce;
  }   

  d->pdata[0]->diversity_score = nbr_diversity_fields; 
 
 /* now do the counting for the rest of the rows */


 for (i = 1 ; i < d->nbr_pdata_row ; i++)
  {
    fc1 = d->pdata[i-1]->fc;

    d->pdata[i]->diversity_score = 0;

    /* copy the linked list for fce's from the previous row */

    for (fc = d->pdata[i]->fc; fc != (FIELD_COUNT *) NULL ; fc = fc->next, fc1 = fc1->next)
     {
       
       fc->Nbr_fce = fc1->Nbr_fce;


       fce2 = (FIELD_COUNT_ELEMENT *) NULL;

       for (fce1 = fc1->fce ; fce1 != (FIELD_COUNT_ELEMENT *) NULL ; fce1 = fce1->next)
        {
         fce = (FIELD_COUNT_ELEMENT *) calloc((size_t) 1, sizeof(FIELD_COUNT_ELEMENT));
         if (fce == (FIELD_COUNT_ELEMENT *) NULL)  
          {
            fprintf(stderr,
	        " *** ERROR:  Out of memory in Diversity_count_diversity_fields_3()\n");
            exit(EXIT_FAILURE);
          }
         fce->next = (FIELD_COUNT_ELEMENT *) NULL;
         fce->nbr_configs = fce1->nbr_configs;
         fce->field_value = fce1->field_value; 
         if (fce2 == (FIELD_COUNT_ELEMENT *) NULL)
          fc->fce = fce;
         else
          fce2->next = fce;
         fce2 = fce;
        }
      } 
        
   if (d->debug)
    {
     printf("i = %d ",i);
    }
   /* do the actual counting */

   for (fc = d->pdata[i]->fc; fc != (FIELD_COUNT *) NULL ;  fc = fc->next)
    {
      for (fce = fc->fce ; fce != (FIELD_COUNT_ELEMENT *) NULL ; fce = fce->next)
       {
	/*found a match for a value previously encountered */

	if (strcmp(fce->field_value,d->pdata[i]->data[fc->fld->col]) == 0)
         {
	  fce->nbr_configs += 1;
	  break;
         }

       }
     
     /* see if a field value not previously encountered */

      if (fce == (FIELD_COUNT_ELEMENT *) NULL)
       {
         
         fce = (FIELD_COUNT_ELEMENT *) calloc((size_t) 1, sizeof(FIELD_COUNT_ELEMENT));
         if (fce == (FIELD_COUNT_ELEMENT *) NULL )   
          {
            fprintf(stderr,
	        " *** ERROR:  Out of memory in Diversity_count_diversity_fields_4()\n");
            exit(EXIT_FAILURE);
          }
        fce->field_value = d->pdata[i]->data[fc->fld->col];
        fce->nbr_configs = 1;
        fc->Nbr_fce += 1;
        fce->next = fc->fce;
        fc->fce = fce; 

       }

    /* calculate the metric */

     metric = (int *) calloc((size_t) fc->Nbr_fce , sizeof(int));
     if (metric == (int *) NULL)    
      {
       fprintf(stderr," *** ERROR:  Out of memory in Diversity_count_diversity_fields_2()\n");
       exit(EXIT_FAILURE);
      }

     if (d->debug)
      {
       printf("(fc->Nbr_fce=%d)",fc->Nbr_fce);
       fflush(stdout);
      }

     for (fce = fc->fce, j=0 ; fce != (FIELD_COUNT_ELEMENT *) NULL && j < fc->Nbr_fce ; 
	fce = fce->next, j++)
      {
        if (fce->nbr_configs > fc->fld->min_nbr_configs_per_option)
	 metric[j] = fc->fld->min_nbr_configs_per_option;
        else
         metric[j] = fce->nbr_configs;

       if (d->debug)
        {
         printf("(metric[%d]=%d)",j,metric[j]);
         fflush(stdout);

        }
      }

     if (d->debug) printf("\n");

     /* sort metric */

     for (j=0; j < fc->Nbr_fce ; j++)
      {
        for (k=j+1 ; k < fc->Nbr_fce ; k++)
         {
           if (metric[k] > metric[j])
            {
             tmp = metric[j];
             metric[j] = metric[k];
             metric[k] = tmp;
            }
          }

       }  

     /* count up to min_nbr_options */

      fc->Nbr_metric = 0;
      
      for (j= 0 ; j < fc->fld->min_nbr_options && j < fc->Nbr_fce ; j++)
       fc->Nbr_metric += metric[j];
           
      free(metric);

      d->pdata[i]->diversity_score += fc->Nbr_metric;
 
     if (d->debug)
      {
       printf("(%d/%d)", fc->Nbr_metric, d->pdata[i]->diversity_score);
       fflush(stdout);
      }       
    

     } 
    
    if (d->debug)
     {
      printf(" d->pdata[%d]->diversity_score = %d\n", i, d->pdata[i]->diversity_score);
      fflush(stdout);

     }       

   }


 if (d->debug != 0)
  {
   printf("\n");
   for (i = 0 ; i < d->nbr_pdata_row ; i++)
    {
 
     for (fc = d->pdata[i]->fc; fc != (FIELD_COUNT *) NULL ;  fc = fc->next)
      {
        printf("row = %d  field = %s  col = %d field_val = %s Nbr_fce = %d\n",
		i, fc->fld->name, fc->fld->col,d->pdata[i]->data[fc->fld->col],fc->Nbr_fce);
	
 	for (fce = fc->fce ; fce != (FIELD_COUNT_ELEMENT *) NULL ; fce = fce->next)
         printf("     Value = %s  count = %d\n",fce->field_value,fce->nbr_configs);
      

      }
  
    }

   fflush(stdout);
  }


 return EXIT_SUCCESS;
}


/* sets the "select" field to 1 up to and including the row that meet criteria */

int Diversity_id_points(DIVERSITY *d) 
{
 int i, cfg_cnt, select;
 FIELD_COUNT *fc;
 FIELD_COUNT_ELEMENT *fce;
 FIELD *div_fld;
 
 if (d->debug) printf("\n");

 select = 1;

 for (i=0 ; i < d->nbr_pdata_row ; i++)
  {
   d->pdata[i]->select = select;

   if (d->pdata[i]->diversity_score == d->metric_base_sum)
    select = 0;
 
   if (d->debug)
    {
     printf("d->pdata[%d]->diversity_score = %d (%d)  select = %d \n",
		i,d->pdata[i]->diversity_score,d->metric_base_sum,d->pdata[i]->select);
    }

  }


 return EXIT_SUCCESS;
}


int Diversity_calculate_statistics(DIVERSITY *d)
{

 double *val;
 double *weight;
 int i,n;

 FIELD *sf;

 if (d == (DIVERSITY *) NULL)
  return EXIT_FAILURE;

 if (d->nbr_pdata_row < 1)
  return EXIT_FAILURE;

 
 val    = (double *)calloc((size_t)d->nbr_pdata_row,sizeof(double));
 weight = (double *)calloc((size_t)d->nbr_pdata_row,sizeof(double));
 
 if (val == (double *) NULL || weight == (double *) NULL)
  {
   fprintf(stderr,
            " *** ERROR:  Out of memory in Diversity_calculate_statistics()\n");
   exit(EXIT_FAILURE);
  }
  
 for (sf = d->statistics_field ; sf != (FIELD *) NULL ; sf = sf->next)
  {
   for (i = 0, n = 0 ; i < d->nbr_pdata_row ; i++)
    {
      if (d->pdata[i]->select == 0)
       continue;   
     
      weight[n] = d->pdata[i]->diversity_score;
      val[n++] = strtod(d->pdata[i]->data[d->statistics_field->col],(char **) NULL);
    }

   sf->st->n = n;
   sf->st->val = val;      /* values not preserved */
   sf->st->weight= weight; /* values not preserved */

   if (calc_stats(sf->st) != EXIT_SUCCESS)
    {
     fprintf(stderr," *** ERROR: in calc_stats()\n");
     return EXIT_FAILURE;
    }
  
  }
 
 free(val);
 free(weight);

 return EXIT_SUCCESS;
}

int Diversity_write_statistics_file(DIVERSITY *d,FILE *sout)
{

 FIELD *st;
 time_t tod;
 char *ctod;
 int i;
 FIELD *df;
 FIELD_COUNT *fc, *fcl;
 FIELD_COUNT_ELEMENT *fce;

 tod = time((time_t *) NULL);
 ctod = ctime(&tod);


 fprintf(sout,"DIVERSITY Version %s (%s)\n",VERSION,__DATE__);
 fprintf(sout,"run time: %s\n",ctod);

 if (d->name != (char *) NULL)
  fprintf(sout,"Name                   = %s\n",d->name);
 
 fprintf(sout,"Input XML file         = %s\n",d->input_xml_filename);
 fprintf(sout,"Input Data File        = %s\n",d->input_data_filename);
 if (d->output_data_filename != (char *) NULL)
  fprintf(sout,"Output Data File       = %s\n",d->output_data_filename);
 fprintf(sout,"Output Statistics File = %s\n",d->output_statistics_filename);


 for (st = d->statistics_field ; st != (FIELD *) NULL ; st = st->next)
  {
   fprintf(sout,"\nFIELD %s (column = %d)\n", st->name, st->col);

   fprintf(sout,"  n            = %d\n",st->st->n);
   fprintf(sout,"  mean         = %f\n",st->st->mean);
   fprintf(sout,"  std dev      = %f\n",st->st->standard_deviation);
   fprintf(sout,"  minimum      = %f\n",st->st->min);
   fprintf(sout,"  5th %%        = %f\n",st->st->percentile_5);
   fprintf(sout,"  median       = %f\n",st->st->median);
   fprintf(sout,"  95th %%       = %f\n",st->st->percentile_95);
   fprintf(sout,"  maximum      = %f\n",st->st->max);
   fprintf(sout,"  weighted avg = %f\n",st->st->weighted_mean); 

  }

 /* check to see if the first row meets the criteria */

 for (fc = d->pdata[0]->fc ; fc != (FIELD_COUNT *) NULL && d->pdata[0]->select == 1 ; 
		fc = fc->next) 
  {
   if (fc->Nbr_metric == fc->fld->metric_base)
    {
     fprintf(sout,"\nFIELD %s: ROW %d (MIN_NBR_OPTON = %d)(MIN_NBR_CONFIGS_PER_OPTION = %d)\n",
		fc->fld->name,0,fc->fld->min_nbr_options, fc->fld->min_nbr_configs_per_option);
      
     for (fce = fc->fce ; fce != (FIELD_COUNT_ELEMENT *)NULL ; fce = fce->next)
      {	
	if (fce->nbr_configs < fc->fld->min_nbr_configs_per_option)
         continue;
         
        fprintf(sout,"   COMPONENT %s (%d configs)\n",fce->field_value,fce->nbr_configs);
      }
    }
  }

 for (i = 1 ; i < d->nbr_pdata_row && d->pdata[i]->select == 1; i++)
  {
   for (fc = d->pdata[i]->fc,fcl = d->pdata[i-1]->fc ; 
        fc != (FIELD_COUNT *) NULL && fcl != (FIELD_COUNT *) NULL ; 
        fc = fc->next, fcl = fcl->next) 
    {

     if (fc->fld != fcl->fld)
      {
       fprintf(stderr," *** ERROR:  fc->fld != fcl->fld in Diversity_write_statistics_file()\n");
       return EXIT_FAILURE;
       
      }
     if (fc->Nbr_metric == fc->fld->metric_base && fcl->Nbr_metric != fcl->fld->metric_base)
      {
       fprintf(sout,"\nFIELD %s: ROW %d (MIN_NBR_OPTON = %d)(MIN_NBR_CONFIGS_PER_OPTION = %d)\n",
	  	  fc->fld->name,i,fc->fld->min_nbr_options, fc->fld->min_nbr_configs_per_option);
      
       for (fce = fc->fce ; fce != (FIELD_COUNT_ELEMENT *)NULL ; fce = fce->next)
        {	
	  if (fce->nbr_configs < fc->fld->min_nbr_configs_per_option)
           continue;
         
          fprintf(sout,"   COMPONENT %s (%d configs)\n",fce->field_value,fce->nbr_configs);
 
        }
      }
    }

  }
 return EXIT_SUCCESS;
}

int Diverstiy_write_output_data_file(DIVERSITY *d,FILE *dout)
{
 FIELD *ofld;
 int i;
 double metric;

 for (ofld = d->output_field ; ofld != (FIELD *) NULL ; ofld = ofld->next)
  {
   fprintf(dout,"\"%s\"",ofld->name);
   if (ofld->next != (FIELD *)NULL) fprintf(dout,",");

  } 

 if (d->output_diversity_metric_flag)
  fprintf(dout,",\"Diversity Metric\"");

 fprintf(dout,"\n");

 for (i = 0 ; i < d->nbr_pdata_row ; i++)
  {
   if (d->pdata[i]->select != 1) continue;

   for (ofld = d->output_field ; ofld != (FIELD *) NULL ; ofld = ofld->next)
    {
     fprintf(dout,"\"%s\"",d->pdata[i]->data[ofld->col]);
     if (ofld->next != (FIELD *)NULL) fprintf(dout,",");
    } 

   if (d->metric_base_sum < 1) metric = 1;
   else metric = (double) d->pdata[i]->diversity_score / (double) d->metric_base_sum; 
    
   if (d->output_diversity_metric_flag)
    fprintf(dout,",%f",metric);

   fprintf(dout,"\n");
    
  }

 fflush(dout);
 return EXIT_SUCCESS;
}

int calc_stats(STATISTICS *s)
{
  int i,j;
  double tmp;

  if (s == (STATISTICS *) NULL) return EXIT_FAILURE;
  if (s->n < 1 || s->val == (double *) NULL) return EXIT_FAILURE;

  /* calculate the mean */

  s->mean = 0.0;

  for (i = 0 ; i < s->n ; i++)
   s->mean += s->val[i];

  s->mean = s->mean / s->n;

 /* calculate the standard deviation using Bessel's correction */

  s->standard_deviation = 0.0;

  if (s->n > 1)
   {
     for (i=0 ; i < s->n ; i++)
      s->standard_deviation += (s->val[i] - s->mean)*(s->val[i] - s->mean);

     s->standard_deviation = sqrt(s->standard_deviation) / (s->n - 1.0); 
   }

  /* calculate the weighted mean if s->weight is non NULL */

  s->weighted_mean = 0.0;
  tmp = 0.0;

  if (s->weight != (double *) NULL)
   {
     for (i=0 ; i < s->n ; i++)
      { 
       s->weighted_mean += s->val[i] * s->weight[i];
       tmp += s->weight[i];
      }

     if (tmp != 0.0)
      s->weighted_mean = s->weighted_mean / tmp;
     else
      s->weighted_mean = 0.0;	
   }

  /* sort val to calculate percentiles and median */

  for (i=0 ; i < s->n ; i++)
   {
     for (j = i+1 ; j < s->n ; j++)
      {
       if (s->val[j] < s->val[i])
        {
         tmp = s->val[i];
         s->val[i] = s->val[j];
         s->val[j] = tmp;
        }
      } 
   }

  s->min = s->val[0];

  s->max = s->val[s->n - 1];
  
  i = (int) floor(0.05 * s->n - 0.5);
  s->percentile_5 = s->val[i];

  i = (int) floor(0.95 * s->n - 0.5);
  s->percentile_95 = s->val[i];

  /* calculate the median based on whether there are an even or odd number of points */

  i = (int) floor(s->n / 2.0); 

  if ((double) i == (double) s->n / 2.0)  /* even */
   {
     s->median = 0.5 * (s->val[i-1] + s->val[i]); 
   }
  else
   {
     s->median = s->val[i];
   }

  return EXIT_SUCCESS;
}

