/* diversity.h */



typedef struct Diversity 
{
  char *input_xml_filename;
  char *input_data_filename;

  char *output_statistics_filename;
  char *output_data_filename;

  char *name;

  int preference; /* PREFERENCE_LOW or PREFERENCE_HIGH */

  int min_nbr_fields;   /* minimum number of fields a row must have to be in pdata */

  struct Data_Row *data;   /* linked list of data,  first element has column labels */
  struct Data_Row **pdata;  /* array of data, eliminating rows with too few fields,
				then sorted based on the sort_field, does not include 
				first row with column labels */
  int nbr_data_row;  /* number of elements in data linked list */
  int nbr_pdata_row;  /* number of elements in pdata -- after ignoring rows with too few fields (min_nbr_fields) */

  int row_dm;  /* first row in sorted pdata for which diversity criteria met */

  int metric_base_sum;  /* sum of the metric base for the field count */

 
  struct Field *sort_field;  /* one field to sort on */
  struct Field *statistics_field;  /* linked List of numeric fields to calculate properties */
  struct Field *output_field;  /* linked list of fields to write to output file */
  struct Field *diversity_field;  /* linked list of fields to base diversity on */

  int output_diversity_metric_flag; /* ON or OFF to create a column for the diversity metric */


  int debug;

}
DIVERSITY;

#define PREFERENCE_LOW 0
#define PREFERENCE_HIGH 1


typedef struct Data_Row
{
  char **data;  /* array of pointers to strings for the fields */

  int n;   /* number of columns */

  struct Data_Row *next;

  struct Field_Count *fc;

  int select; /* = 0 for no, = 1 for yes,  row is selected for statistics, and output */

  int diversity_score;

}
DATA_ROW;


typedef struct Field
{
  char *name;
  int  col;  /* start with 0 */
  struct Field *next;

  /* following only used for diversity fields */
  
  int min_nbr_options;
  int min_nbr_configs_per_option;
  int metric_base;  /* minimum number of components times minimum number of configs */
  

  /* following only used for statistics */

  struct Statistics *st;

}
FIELD;


typedef struct Field_Count
{
 struct Field *fld;   /* correlates to which field (from diversity_field) we are looking at */

 struct Field_Count_Element *fce; 
	/* linked list: holds count of number of configurations encountered so far for each field value */

 int Nbr_fce; /* number of distinct field values encountered so far */

 int Nbr_metric;   /* number of configs (up to min number) for up to min number of components */

 
struct Field_Count *next;
  
}
FIELD_COUNT;

typedef struct Field_Count_Element
{
 char *field_value;
 int  nbr_configs;  
 struct Field_Count_Element *next;
}
FIELD_COUNT_ELEMENT;


typedef struct Statistics
{
 double mean;
 double median;
 double standard_deviation;
 double percentile_5;
 double percentile_95; 
 double weighted_mean;
 double min;
 double max;

 int n;
 double *val;
 double *weight;

}
STATISTICS;


int Diversity_initialize(DIVERSITY *);
int Diversity_read_command_line(DIVERSITY *,int, char **);
int Diversity_read_input_xml_file(DIVERSITY *, FILE *);
int Diversity_read_input_csv_file(DIVERSITY *, FILE *);
int Diversity_sort_data(DIVERSITY *);
int Diversity_count_diversity_fields(DIVERSITY *);
int Diversity_id_points(DIVERSITY *);
int Diversity_calculate_statistics(DIVERSITY *);
int Diversity_write_statistics_file(DIVERSITY *,FILE *);
int Diverstiy_write_output_data_file(DIVERSITY *,FILE *);


DATA_ROW *read_csv_file(FILE *in, int *ncnt);
char *copystring(char *s);
