/* readcsv.c */

/* (C) Norbert Doerry 2014
   November 2014


This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.



*/


#include <stdio.h>
#include <stdlib.h>
#include "diversity.h"

char *strcpy(char *dest, const char *src);
void strstriptrail(char *s);
size_t strlen(const char *s);


DATA_ROW *parse_csv_line(char *s);

#define MAX_NBR_FIELDS 5000
#define MAX_BUF_SIZE 100000

DATA_ROW *read_csv_file(FILE *in, int *ncnt)
{

 DATA_ROW *base,*last,*dr;
 char *buf;


 base = (DATA_ROW *) NULL;
 last = (DATA_ROW *) NULL;

 if (ncnt != (int *) NULL)
  *ncnt = 0;

 buf = (char *) calloc((size_t) MAX_BUF_SIZE,sizeof(char));
 
 if (buf == (char *) NULL)
  {
   fprintf(stderr," *** ERROR: Out of memory in read_csv_file() buf\n");
   exit(EXIT_FAILURE);
  }

 while (fgets(buf,MAX_BUF_SIZE,in) > 0)
  {
    dr = parse_csv_line(buf);

    if (base == (DATA_ROW *) NULL)   /* attach to linked list of data rows */
     { 
      base = dr;
     }
    else
     {
      last->next = dr;
     }

    last = dr;

    if (ncnt != (int *) NULL) *ncnt += 1;  /* update the count of lines read in */
 
  }

 free(buf); 

 return base;

}


/* conversion rules ...


a field ends with a comma
leading white spaces are ignored
if first character is a ", then the last character should also be a "  the field is everything in between
including commas. 
if first character is a ", and  "" is encountered following the ", they are part of the field and not the ending "
Trailing white spaces and any characters (other than comma) after a closing " are ignored.
*/

DATA_ROW *parse_csv_line(char *s)
 {

  int istart, istop, i;
  DATA_ROW *dr;
  char *data[MAX_NBR_FIELDS];
  char c;

  if (s == (char *) NULL) return (DATA_ROW *) NULL;

  istart = 0;
  istop = 0;
  dr = (DATA_ROW *) NULL;

  while (s[istart] != (char) NULL)
   {

    while (isspace(s[istart]))   /* ignore leading white spaces */
     istart++;

    if (s[istart] == (char) NULL) break;  /* see if we are at the end of the string */

    if (dr == (DATA_ROW *) NULL)  /* allocate dr if it doesn't already exist */
     {

      dr = (DATA_ROW *) calloc((size_t) 1, sizeof(DATA_ROW));
      if (dr == (DATA_ROW *) NULL)
       {
        fprintf(stderr," *** ERROR:  Out of memory in parse_csv_line() (dr)\n");
	exit(EXIT_FAILURE);

       }

      dr->data = (char **) NULL;
      dr->n = 0;
      dr->next = (DATA_ROW *) NULL;
     }
    istop = istart;

    if (s[istart] == '"')  /* field is a text field delimited with "*/
     {
      for (istop = istart + 1 ; istop != (char) NULL ; istop++)
       {
        if (s[istop] == '"')
         {
           istop++;

	   if (s[istop] != '"')
            {
             break;  /* end of field */
	    }
         }
      

       } 



     }
    else
     {

      for (istop = istart ; s[istop] != (char) NULL && s[istop] != ','; istop++);

     }    


    
    /* allocate and store the field that begins with istart and ends with istop-1 */
    
    c = s[istop];
    s[istop] = (char) NULL;

    data[dr->n] = copystring((char *)(s+istart));
    strstriptrail(data[dr->n]);
    dr->n += 1;
    s[istop] = c;

 
    /* update istop for text fields */

     while (s[istop] != (char) NULL && s[istop] != ',')istop++;  /* ignore trailing characters in field */

    /* update istart for the next field */

    if (s[istop] != (char) NULL) 
	istart = istop +1;
    else
	istart = istop;

   }  


  if (dr->n < 1)
   return dr;


  dr->data = (char **)calloc(dr->n,sizeof(char *));

  if (dr->data == (char **) NULL)
   {
    fprintf(stderr," *** ERROR:  Out of memory in parse_csv_line() (dr->data)\n");
    exit(EXIT_FAILURE);
   }

  for (i = 0 ; i < dr->n ; i++)
   dr->data[i] = data[i];
  
  return dr;
 }





char *copystring(char *s)
{
 int i;
 char *ps;

 if (s == (char *) NULL)
  return s;

 ps = (char *) calloc(strlen(s)+1,sizeof(char));

 if (ps == (char *) NULL)
  {
   fprintf(stderr," *** ERROR:  Out of memory in copystring()\n");
   exit(EXIT_FAILURE);
  }
 
 strcpy(ps,s);

 return ps; 

}

void strstriptrail(char *s)
{
 int i;

 if (s == (char *) NULL) return;

 i = strlen(s)-1;
 
 while (isspace(s[i]) && i >= 0)
  {
   s[i] = (char) NULL;
   i--;
  }
 
 return;
}


